diff --git a/src/plugins/quota/quota-storage.c b/src/plugins/quota/quota-storage.c
index 4c0bdcf..522ac97 100644
--- a/src/plugins/quota/quota-storage.c
+++ b/src/plugins/quota/quota-storage.c
@@ -185,14 +185,29 @@ void quota_mail_allocated(struct mail *_mail)
 	MODULE_CONTEXT_SET_SELF(mail, quota_mail_module, qmail);
 }
 
-static int quota_check(struct mail_save_context *ctx)
+static bool quota_roots_equal(struct mailbox *box1, struct mailbox *box2)
+{
+	struct quota_user *quser = QUOTA_USER_CONTEXT(box1->list->ns->user);
+	struct quota_root *const *rootp;
+
+	array_foreach(&quser->quota->roots, rootp) {
+		if (quota_root_is_namespace_visible(*rootp, box1->list->ns) !=
+		    quota_root_is_namespace_visible(*rootp, box2->list->ns))
+			return FALSE;
+	}
+	return TRUE;
+}
+
+static int quota_check(struct mail_save_context *ctx, struct mailbox *src_box)
 {
 	struct mailbox_transaction_context *t = ctx->transaction;
 	struct quota_transaction_context *qt = QUOTA_CONTEXT(t);
 	int ret;
 	bool too_large;
 
-	if (ctx->moving) {
+	if (ctx->moving &&
+	    (src_box == NULL ||
+	     quota_roots_equal(ctx->transaction->box, src_box))) {
 		/* the mail is being moved. the quota won't increase (after
 		   the following expunge), so allow this even if user is
 		   currently over quota */
@@ -242,7 +257,7 @@ quota_copy(struct mail_save_context *ctx, struct mail *mail)
 		   quota */
 		return 0;
 	}
-	return quota_check(ctx);
+	return quota_check(ctx, mail->box);
 }
 
 static int
@@ -301,7 +316,7 @@ static int quota_save_finish(struct mail_save_context *ctx)
 	if (qbox->module_ctx.super.save_finish(ctx) < 0)
 		return -1;
 
-	return quota_check(ctx);
+	return quota_check(ctx, NULL);
 }
 
 static void quota_mailbox_sync_cleanup(struct quota_mailbox *qbox)
